/******************************************************************************
 *
 *  INPUT HANDLERS
 *
 *  Left mouse button:   start training session 
 *  Right mouse button:  start testing session
 *  'B'-key:             increase training blocksize
 *  'b'-key:             decrease training blocksize 
 *  'f'/'F'-key:         toggle between fast and slow training
 *  't'/'T'-key:         toggle language
 *  '+'-key:             increase frameRate for auto testing
 *  '-'-key:             decrease frameRate for auto testing
 *  'h'/'H'-key:         hide / show actual digit
 *
 ******************************************************************************/

/******************************************************************************
 *
 *  MOUSE HANDLER
 *
 ******************************************************************************/
void mousePressed() 
{
  if (help)
  {  
    help = !help;
    return;
  }

  if (mouseButton == LEFT)
  {
    // LEFT MOUSE BUTTON PRESSED: TRAINING
    if (useAutoTesting) frameRate(FRAMERATE_DEFAULT);

    if (useFastTraining)
    { // FAST TRAINING WITHOUT DISPLAYING
      fastTraining(trainingBlocksize);
      nrOfTestedImages   = 0;
      nrOfMatchedImages  = 0;
      networkJustTrained = true;
    } else
    { // SLOW (ANIMATED) TRAINING
      nrOfTestedImages   = 0;
      nrOfMatchedImages  = 0;
      fSuccessPercentage = 0.0;
      fLastAccuracy      = 0.0;
      // NUMBER OF IMAGES FOR TRAINING      
      trainingImagesLeft = trainingBlocksize;
    }
  } else
  { 
    // RIGHT MOUSE BUTTON PRESSED: TESTING
    useAutoTesting = !useAutoTesting;
    if (useAutoTesting)
      frameRate(frameRateAutoTesting);
    else
      frameRate(FRAMERATE_DEFAULT);
    // println("useAutoTesting: "+useAutoTesting);
  }
} // mousePressed()


/******************************************************************************
 *
 *  KEYBOARD HANDLER
 *
 ******************************************************************************/
void keyPressed()
{
  if (help)
  {  
    help = !help;
    return;
  }  

  if (key == ' ')
  {
    // TEST THE NEXT NEW IMAGE
    testingNextImage = true;
  } else if (key == 'f' || key == 'F')
  {
    // TOGGLE BETWEEN FAST- AND SLOW TRAINING
    useFastTraining = !useFastTraining;
    println("useFastTraining: "+useFastTraining);
  } else if (key == 'h' || key == 'H' || key == '?' || key == '/')
  {
    // TOGGLE HELP SCREEN
    help = !help;
    // REDRAW SCREEN
    displayCanvas();
    loop();
  } else if (key == 'l')
  {
    // DECREASE NUMBER OF HIDDEN LAYERS
    if (NR_OF_HIDDEN_LAYERS > 0)
    {  
      NR_OF_HIDDEN_LAYERS--;
      initSettings();
    }
  } else if (key == 'L')
  {
    // INCREASE NUMBER OF HIDDEN LAYERS
    if (NR_OF_HIDDEN_LAYERS < 5)
    {  
      NR_OF_HIDDEN_LAYERS++;
      initSettings();
    }
  } else if (key == 'n')
  {
    // DECREASE NUMBER OF NEURONS IN THE HIDDEN LAYERS
    if (CURRENT_NR_OF_HIDDEN_NEURONS_INDEX > 0)
    {  
      CURRENT_NR_OF_HIDDEN_NEURONS_INDEX--;
      initSettings();
    }
  } else if (key == 'N')
  {
    // DECREASE NUMBER OF NEURONS IN THE HIDDEN LAYERS
    if (CURRENT_NR_OF_HIDDEN_NEURONS_INDEX < POSSIBLE_HIDDEN_NEURONS.length-1)
    {  
      CURRENT_NR_OF_HIDDEN_NEURONS_INDEX++;
      initSettings();
    }
  } else if (key == 't' || key == 'T')
  {
    // TOGGLE LANGUAGE (EN <> NL)
    nl = !nl;
  } else if (key == '+' || key == '=')
  {
    // INCREASE AUTO TESTING SPEED
    frameRateAutoTesting++;
    if (useAutoTesting)
      frameRate(frameRateAutoTesting);
    else
      frameRate(FRAMERATE_DEFAULT);
    println("frameRateAutoTesting: "+frameRateAutoTesting);
  } else if (key == '-' || key == '_')
  {
    // DECREASE AUTO TESTING SPEED
    frameRateAutoTesting--;
    if (useAutoTesting)
      frameRate(frameRateAutoTesting);
    else
      frameRate(FRAMERATE_DEFAULT);
    println("frameRateAutoTesting: "+frameRateAutoTesting);
  } else if (key == 'b')
  {  // DECREASE TRAINING BLOCKSIZE
    if (trainingBlocksize > 100) trainingBlocksize -= 100;
    println("trainingBlocksize: "+trainingBlocksize);
  } else if (key == 'B')
  {  // INCREASE TRAINING BLOCKSIZE
    trainingBlocksize += 100;
    println("trainingBlocksize: "+trainingBlocksize);
  }
} // keyPressed()