/******************************************************************************
 *
 *  LOAD MNIST TRAINING- AND TEST-DATA
 *
 *  http://yann.lecun.com/exdb/mnist/
 *
 ******************************************************************************/

DatasetRecord [] trainingSet;
DatasetRecord [] testingSet;

/******************************************************************************
 *
 *  DATASET RECORD CLASS
 *
 ******************************************************************************/
class DatasetRecord
{
  /****************************************************************************
   *
   *  PROPERTIES
   *
   ****************************************************************************/
  // PIXEL VALUES OF THE IMAGE (-1.0 = WHITE; +1.0 = BLACK)
  float [] inputs;
  // +1.0 INDICATES THE CORRECT ANSWER (THE REST IS -1.0)
  float [] outputs;
  int output;


  /****************************************************************************
   *
   *  CONSTRUCTOR
   *
   ****************************************************************************/
  DatasetRecord()
  {
    inputs  = new float [NR_OF_INPUT_NEURONS];
    outputs = new float [NR_OF_OUTPUT_NEURONS];
  } // DatasetRecord()


  /****************************************************************************
   *
   *  LOAD IMAGES
   *
   ****************************************************************************/
  void imageLoad(byte [] images, int offset)
  {
    for (int i = 0; i < NR_OF_INPUT_NEURONS; i++)
      inputs[i] = int(images[i+offset]) / 128.0 - 1.0;
  } // imageLoad()


  /****************************************************************************
   *
   *  LOAD LABELS
   *
   *  LABELS ARE THE CORRECT ANSWERS, IN THIS CASE '0'..'9'
   *
   ****************************************************************************/
  void labelLoad(byte [] labels, int offset)
  {
    output = int(labels[offset]);
    for (int i = 0; i < NR_OF_OUTPUT_NEURONS; i++)
      if (i == output)
        outputs[i] = 1.0;
      else
        outputs[i] = -1.0;
  } // labelLoad()
} // DatasetRecord


/******************************************************************************
 *
 *  CREATE AND FILL DATASET OBJECTS
 *
 ******************************************************************************/
void loadData()
{
  // THE IMAGES (196 BYTES PER IMAGE)
  byte [] images = loadBytes("t10k-images-14x14.idx3-ubyte");
  // THE CORRECT OUTPUTS (0..9)
  byte [] labels = loadBytes("t10k-labels.idx1-ubyte");

  // TRAINING IMAGES
  trainingSet = new DatasetRecord [TRAINING_SET_SIZE];
  int trainPos = 0;

  // TESTING IMAGES
  testingSet = new DatasetRecord [TESTING_SET_SIZE];
  int testPos = 0;

  for (int i = 0; i < (TRAINING_SET_SIZE + TESTING_SET_SIZE); i++)
  {
    if (i % 5 != 0)
    {
      // ADD TRAINING DATA RECORD (4/5th OF THE TOTAL AMOUNT OF RECORDS)
      trainingSet[trainPos] = new DatasetRecord();
      trainingSet[trainPos].imageLoad(images, 16 + i * NR_OF_INPUT_NEURONS);
      trainingSet[trainPos].labelLoad(labels, 8 + i);
      trainPos++;
    } else
    {
      // ADD TESTING DATA RECORD (1/5th OF THE TOTAL AMOUNT OF RECORDS)
      testingSet[testPos] = new DatasetRecord();
      testingSet[testPos].imageLoad(images, 16 + i * NR_OF_INPUT_NEURONS);
      testingSet[testPos].labelLoad(labels, 8 + i);
      testPos++;
    }
  }
} // loadData()