/******************************************************************************
 *
 *  NEURON CLASS
 *
 ******************************************************************************/
float LEARNING_RATE = 0.01;

class Neuron
{
  Neuron [] inputNeurons;
  float [] neuronWeights;
  float neuronTreshold = 0.0;
  float neuronOutput   = 0.0;
  float neuronError;


  /****************************************************************************
   *
   *  CONSTRUCTOR FOR THE INPUT LAYER
   *
   ****************************************************************************/
  Neuron()
  {
    // INITIAL RANDOM OUTPUT
    neuronOutput = sigmoid.lookupSigmoid(random(-5.0, 5.0));
  } // Neuron()


  /****************************************************************************
   *
   *  CONSTRUCTOR FOR THE HIDDEN- AND OUTPUT LAYER
   *
   ****************************************************************************/
  Neuron(Neuron[] inputs)
  {
    inputNeurons  = new Neuron [inputs.length];
    neuronWeights = new float [inputs.length];

    // CREATE RANDOM WEIGHTS
    for (int i = 0; i < inputs.length; i++)
    {
      inputNeurons[i] = inputs[i];
      neuronWeights[i] = random(-1.0, 1.0);
    }

    // RANDOM TRESHOLD
    neuronTreshold = random(-1.0, 1.0);
    neuronError = 0.0;

    // INITIAL RANDOM OUTPUT
    neuronOutput = sigmoid.lookupSigmoid(random(-5.0, 5.0));
  } // Neuron()


  /****************************************************************************
   *
   *  RESPOND LOOKS AT THE LAYER BELOW, AND PREPARES A RESPONSE
   *
   ****************************************************************************/
  float respond()
  {
    float input = 0.0;
    for (int i = 0; i < inputNeurons.length; i++) {
      input += inputNeurons[i].neuronOutput * neuronWeights[i];
    }
    neuronOutput = sigmoid.lookupSigmoid(input + neuronTreshold);
    // RESET OUR ERROR VALUE READY FOR TRAINING
    neuronError = 0.0;

    return neuronOutput;
  } // respond()


  /****************************************************************************
   *
   *  FIND ERROR IS USED ON THE OUTPUT NEURONS
   *
   ****************************************************************************/
  void finderror(float desired)
  {
    neuronError = desired - neuronOutput;
  } // finderror()


  /****************************************************************************
   *
   *  TRAIN ADJUSTS THE WEIGHTS BY COMPARING ACTUAL OUTPUT TO CORRECT OUTPUT
   *
   ****************************************************************************/
  void train()
  {
    float delta = (1.0 - neuronOutput) * (1.0 + neuronOutput) * neuronError * LEARNING_RATE;
    for (int i = 0; i < inputNeurons.length; i++)
    {
      // TELL THE NEXT LAYER DOWN WHAT IT'S DOING WRONG
      inputNeurons[i].neuronError += neuronWeights[i] * neuronError;
      // CORRECT OUR WEIGHTS
      neuronWeights[i] += inputNeurons[i].neuronOutput * delta;
    }
  } // train()


  /****************************************************************************
   *
   *  DRAW THE NEURONS FOR ALL LAYERS
   *
   *  not_hidden: in- or output layer
   *  type: "i", "h" or "o" (input, hidden, output)
   *  nr: sequence number of the neuron
   *
   ****************************************************************************/
  void draw(String type, int nr)
  {
    float level = (0.5-neuronOutput*0.5);

    if (type != "h")
    { // INPUT- OR OUTPUT LAYER
      fill(255 * level);
      if (type == "o")
      { // OUPUT LAYER
        // PRINTS THE NUMBERS ABOUT THE OUTPUT DOTS
        pushStyle();
        fill(0);
        textSize(12);
        text(nr, 15, -6);
        popStyle();
      }
    } else {
      // MERGE HIDDEN LAYER WITH BACKGROUND COLOR
      fill(110 + 128 * level, 102 + 128 * level, 127 + 128 * level);
    }

    // DRAW THE DOT
    if (type == "i")
      // INPUT
      rect(0, 0, INPUT_SQUARE_SIZE, INPUT_SQUARE_SIZE);
    else if (type  == "h")
      // HIDDEN
      ellipse(0, 0, HIDDEN_DOT_SIZE, HIDDEN_DOT_SIZE);
    else
      // OUTPUT
      ellipse(0, 0, OUTPUT_DOT_SIZE, OUTPUT_DOT_SIZE);
  } // draw()
} // Neuron