import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class rvg_neural_net20 extends PApplet {

/******************************************************************************
 *
 *  A SIMPLE NEURAL NETWORK v2.0
 *
 *  v12/05/2015  (c) Rolf van Gelder, Eindhoven, NL - http://cagewebdev.com
 *
 ******************************************************************************/

// THE NETWORK
Network neuralnet;

// SIGMOID LOOKUP TABLE
Sigmoid sigmoid;


/******************************************************************************
 *
 *  SETUP
 *
 ******************************************************************************/
public void setup()
{
  

  // INITIALIZE THE SETTINGS
  initSettings();

  // LOAD TRAINING- AND TEST-DATASETS
  loadData();

  // LAYOUT DEFAULTS
  background(BACKGROUND_COLOR);
  noStroke();
  ellipseMode(CORNER);
  textSize(15);
  textAlign(RIGHT);

  if (useAutoTesting)
    frameRate(frameRateAutoTesting);
  else
    frameRate(FRAMERATE_DEFAULT);
} // setup()


/******************************************************************************
 *
 *  DRAW
 *
 ******************************************************************************/
public void draw()
{
  if (help)
  {  
    helpScreen();
  } else
  {
    if (trainingImagesLeft > 0)
    { // TRAINING SESSION IS RUNNING

      running = true;

      // GET NEXT TRAINING IMAGE
      trainOneImage();

      // ONE LESS TO TRAIN...
      trainingImagesLeft--;

      if (trainingImagesLeft < 1)
      { // LAST IMAGE PROCESSED: RESET COUNTERS
        nrOfTestedImages  = 0;
        nrOfMatchedImages = 0;
        if (useAutoTesting) frameRate(frameRateAutoTesting);
      }
    } else
    {
      running = false;
    } // if (trainingImagesLeft > 0)

    // TEST THE NEXT IMAGE
    if (useAutoTesting) testingNextImage = true;

    if (testingNextImage || networkJustTrained)
    { 
      if (testingNextImage)
      { 
        running = true;
        // TEST A RANDOM TEST IMAGE
        int row = (int) floor(random(0, testingSet.length));
        response = neuralnet.respond(testingSet[row].inputs);
        actual   = testingSet[row].output;
      }
    } else
      running = false;
    // if (testingNextImage || networkJustTrained)

    // DRAW THE NETWORK LAYERS AND STATISTICS
    displayCanvas();
    loop();
  } // if (help)
} // draw()


/******************************************************************************
 *
 *  DRAW SOME STATISTICS
 *
 ******************************************************************************/
public void drawStats()
{
  int margin  = 10;
  int spacing = 24;
  int top     = height-INPUT_SIZE-5;

  // DRAW CONNECTIONS BETWEEN THE LAYERS
  drawConnections();

  // DRAW SOME STATISTICS
  fill(0);
  textAlign(LEFT);
  if (trainingImagesLeft > 0)
  {
    fill(255);
    if (nl)
      rect(margin - 3, 18, 145, 19);
    else
      rect(margin - 3, 18, 155, 19);

    fill(0);
    if (nl)
      text("NETWERK TRAINEN", margin, 33);
    else
      text("TRAINING NETWORK", margin, 33);
  } else if (testingNextImage)
  {
    fill(255);
    if (nl)
      rect(margin - 3, 18, 135, 19);
    else
      rect(margin - 3, 18, 145, 19);

    fill(0);
    if (nl)
      text("NETWERK TESTEN", margin, 33);
    else
      text("TESTING NETWORK", margin, 33);    
    if (response == actual) nrOfMatchedImages++;
    nrOfTestedImages++;
  }
  if (nl)
  {
    text("getrainde afbeeldingen: "+nrOfTrainedImages, margin, top);
    text("geteste afbeeldingen: "+nrOfTestedImages, margin, top+spacing);
    text("treffers: "+nrOfMatchedImages, margin, top+2*spacing);
  } else
  {  
    text("trained images: "+nrOfTrainedImages, margin, top);
    text("tested images: "+nrOfTestedImages, margin, top+spacing);
    text("hits: "+nrOfMatchedImages, margin, top+2*spacing);
  }
  fSuccessPercentage = 0.0f;
  if (nrOfTestedImages > 0) fSuccessPercentage = (float)nrOfMatchedImages*100/nrOfTestedImages;
  if ((nrOfTestedImages%CHECKPOINT_AUTO_TESTING) == 0 && useAutoTesting)
  {
    if (nl)
      println("getrainde afbeeldingen: " +nrOfTrainedImages+" geteste afbeeldingen: "+CHECKPOINT_AUTO_TESTING+" succes: "+fSuccessPercentage+"%");
    else
      println("trained images: " +nrOfTrainedImages+" tested images: "+CHECKPOINT_AUTO_TESTING+" success: "+fSuccessPercentage+"%");
    useAutoTesting = false;
  }
  if (nl)
  {
    text("gemiddeld succes: "+nfc(fSuccessPercentage, 1)+"%", margin, top+3*spacing);
    text("betrouwbaarheid (laatste test): "+nfc(fLastAccuracy, 1)+"%", margin, top+4*spacing);
  } else
  {
    text("average success: "+nfc(fSuccessPercentage, 1)+"%", margin, top+3*spacing);
    text("accuracy (last test): "+nfc(fLastAccuracy, 1)+"%", margin, top+4*spacing);
  }
  pushStyle();
  textAlign(CENTER);
  String layers = "";
  if (NR_OF_HIDDEN_LAYERS > 0)
  {  
    if (NR_OF_HIDDEN_LAYERS > 1)
    {  
      if (nl)
        layers = "verborgen lagen";
      else
        layers = "hidden layer";
    } else
    {  
      if (nl)
        layers = "verborgen laag";
      else
        layers = "hidden layer";
    }
  }
  for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
  {
    if (nl)
      text(layers+" ("+NR_OF_HIDDEN_NEURONS.get(i)+" neuronen)", width>>1, hidden_offset_y[i] - 10);
    else
      text(layers+" ("+NR_OF_HIDDEN_NEURONS.get(i)+" neurons)", width>>1, hidden_offset_y[i] - 10);
  }
  popStyle();

  textAlign(LEFT);
  if (nl)
    text("type 'h' voor hulp", margin, height-25);
  else
    text("press 'h' for help", margin, height-25);

  textAlign(RIGHT);
  text("fps: "+nfc(frameRate, 2), width-margin, height-25);

  // ACTUAL AND RESPONSE
  if (trainingImagesLeft < 1 && nrOfTestedImages > 1)
  {
    int bgcolor, fgcolor;
    if (response != actual)
    {  
      bgcolor = color(255, 0, 0);
      fgcolor = color(255);
    } else
    {  
      bgcolor = color(0, 255, 0);
      fgcolor = color(0);
    }
    fill(bgcolor);
    if (nl)
      rect(width-margin-94, 20, 99, 20);
    else
      rect(width-margin-90, 20, 95, 20);
    fill(fgcolor);
    if (nl)
      text("antwoord: "+str(response), width-margin, 35);
    else
      text("response: "+str(response), width-margin, 35);
    if (!hideActual)
    {
      fill(bgcolor);
      if (nl)
        rect(width-margin-88, input_offset_y-2, 92, 20);
      else
        rect(width-margin-68, input_offset_y-2, 72, 20);
      fill(fgcolor);
      if (nl)
        text("werkelijk: "+str(actual), width-margin, input_offset_y+12);
      else
        text("actual: "+str(actual), width-margin, input_offset_y+12);
    }
  }

  // STOP THE DISPLAY LOOP
  testingNextImage   = false;
  networkJustTrained = false;
} // drawStats()


/******************************************************************************
 *
 *  DRAW CONNECTIONS BETWEEN THE LAYERS
 *
 ******************************************************************************/
public void drawConnections()
{
  // LET THE OUTPUT CONNECTOR POINT TO THE STRONGEST VALUE
  // neuralnet.lastResponse: 0..9 => 1..10
  int r = neuralnet.lastResponse;
  if (r == 0) r = 10;
  int outputX = PApplet.parseInt(outputLeftBottom.x + OUTPUT_DOT_SIZE/2 + (r-1)*(OUTPUT_DOT_SIZE+8));

  pushStyle();

  // TOGGLE THE COLOR OF THE CONNECTION
  stroke(currentColor);
  if (running)
  {
    if (currentColor == color1)
      currentColor = color2;
    else
      currentColor = color1;
  }

  strokeWeight(3);

  if (NR_OF_HIDDEN_LAYERS == 0)
  {
    // CONNECT INPUT DIRECTLY WITH OUTPUT
    line(inputCenterTop.x, inputCenterTop.y, outputCenterBottom.x, outputCenterBottom.y);
  } else
  {
    // CONNECT HIDDEN LAYERS
    for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
    {  
      if (i == 0)
        // FIRST HIDDEN LAYER: CONNECT TO INPUTS
        line(inputCenterTop.x, inputCenterTop.y, hiddenCenterBottom[0].x, hiddenCenterBottom[0].y);
      if (i > 0)
        // MIDDLE HIDDEN LAYER: CONNECT TO PREVIOUS HIDDEN LAYER
        line(hiddenCenterLeft[i].x, hiddenCenterLeft[i].y, hiddenCenterRight[i-1].x, hiddenCenterRight[i-1].y);
      if (i == NR_OF_HIDDEN_LAYERS-1)
        // LAST HIDDEN LAYER: CONNECT TO OUTPUTS
        line(hiddenCenterTop[i].x, hiddenCenterTop[i].y, outputX, outputLeftBottom.y-OUTPUT_DOT_SIZE/2);
    } // for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
  } // if (NR_OF_HIDDEN_LAYERS == 0)

  popStyle();
} // drawConnections()


/******************************************************************************
 *
 *  DRAW THE NETWORK LAYERS AND STATISTICS
 *
 ******************************************************************************/
public void displayCanvas()
{
  background(BACKGROUND_COLOR);

  // CALCULATE AND DRAW THE NETWORK LAYERS 
  neuralnet.draw();

  // DRAW SOME STATISTICS
  drawStats();
} // displayCanvas()


/******************************************************************************
 *
 *  FAST TRAINING WITHOUT DISPLAYING
 *
 ******************************************************************************/
public void fastTraining(int cnt)
{
  // TRAIN THE NETWORK WITH A SET OF IMAGES
  for (int i = 0; i < cnt; i++) trainOneImage();
} // fastTraining()


/******************************************************************************
 *
 *  TRAIN THE NETWORK WITH ONE RANDOM IMAGE
 *
 ******************************************************************************/
public void trainOneImage()
{
  // SELECT A RANDOM TRAINING INPUT AND TRAIN
  int i = (int)floor(random(0, trainingSet.length));
  // WHAT THE NETWORK THINKS IT IS
  response = neuralnet.respond(trainingSet[i].inputs);
  // WHAT IT ACTUALLY IS
  actual   = trainingSet[i].output;
  neuralnet.train(trainingSet[i].outputs);
  nrOfTrainedImages++;
} // trainOneImage()
/******************************************************************************
 *
 *  SHOW HELPSCREEN
 *
 ******************************************************************************/
public void helpScreen()
{
  int mouseYPos    = 350;
  int keyboardYPos = 470;
  int vSpacing     = 30;

  pushStyle();

  background(BACKGROUND_COLOR);
  fill(0);
  textAlign(CENTER);
  textSize(24);
  if (nl)
    text("Een eenvoudig Neuraal Netwerk, door Rolf van Gelder - http://cagewebdev.com", width>>1, 50);
  else
    text("A simple Neural Network, by Rolf van Gelder - http://cagewebdev.com", width>>1, 50);

  image(numberImg, (width>>1)-(numberImg.width>>1), 80);

  textSize(22);
  if (nl)
  {
    text("MUIS", width>>1, mouseYPos);
    text("TOETSENBORD", width>>1, keyboardYPos);
  } else
  {
    text("MOUSE", width>>1, mouseYPos);
    text("KEYBOARD", width>>1, keyboardYPos);
  }
  textSize(16);
  if (nl)
  {
    text("<LINKERMUISKNOP> start een TRAININGSSESSIE", width>>1, mouseYPos+vSpacing);
    text("<RECHTERMUISKNOP> start een TESTSESSIE", width>>1, mouseYPos+2*vSpacing);
    text("<SPATIEBALK> test een volgende afbeelding", width>>1, keyboardYPos+vSpacing);
    text("<b> verlaag het aantal trainingsafbeeldingen per sessie", width>>1, keyboardYPos+2*vSpacing);
    text("<B> verhoog het aantal trainingsafbeeldingen per sessie", width>>1, keyboardYPos+3*vSpacing);
    text("<f> wissel tussen langzame (geanimeerde) en snelle (niet-geanimeerde) training", width>>1, keyboardYPos+4*vSpacing);
    text("<h> toon / verberg dit helpscherm", width>>1, keyboardYPos+5*vSpacing);
    text("<l> verlaag het aantal verborgen lagen (minimaal 0)", width>>1, keyboardYPos+6*vSpacing);
    text("<L> verhoog het aantal verborgen lagen (maximaal 5)", width>>1, keyboardYPos+7*vSpacing);
    text("<n> verlaag het aantal neuronen in de verborgen lagen (minimaal 25)", width>>1, keyboardYPos+8*vSpacing);
    text("<N> verhoog het aantal neuronen in de verborgen lagen (maximaal 144)", width>>1, keyboardYPos+9*vSpacing);
    text("<t> wissel de taal (Nederlands <> Engels)", width>>1, keyboardYPos+10*vSpacing);
  } else
  {    
    text("<LEFT-CLICK> start a TRAINING SESSION", width>>1, mouseYPos+vSpacing);
    text("<RIGHT-CLICK> start a TEST SESSION", width>>1, mouseYPos+2*vSpacing);
    text("<SPACEBAR> test the next image", width>>1, keyboardYPos+vSpacing);
    text("<b> decrease training batch size", width>>1, keyboardYPos+2*vSpacing);
    text("<B> increase training batch size", width>>1, keyboardYPos+3*vSpacing);
    text("<f> toggle between slow (animated) and fast (non-animated) training", width>>1, keyboardYPos+4*vSpacing);
    text("<h> show / hide this help screen", width>>1, keyboardYPos+5*vSpacing);    
    text("<l> decrease the number of hidden layers (min = 0)", width>>1, keyboardYPos+6*vSpacing);
    text("<L> increase the number of hidden layers (max = 5)", width>>1, keyboardYPos+7*vSpacing);
    text("<n> decrease the number of neurons in the hidden layers (min = 25)", width>>1, keyboardYPos+8*vSpacing);
    text("<N> increase the number of neurons in the hidden layers (max = 144)", width>>1, keyboardYPos+9*vSpacing);   
    text("<t> toggle the language (Dutch <> English)", width>>1, keyboardYPos+10*vSpacing);
  }

  popStyle();
} // helpScreen()
/******************************************************************************
 *
 *  INPUT HANDLERS
 *
 *  Left mouse button:   start training session 
 *  Right mouse button:  start testing session
 *  'B'-key:             increase training blocksize
 *  'b'-key:             decrease training blocksize 
 *  'f'/'F'-key:         toggle between fast and slow training
 *  't'/'T'-key:         toggle language
 *  '+'-key:             increase frameRate for auto testing
 *  '-'-key:             decrease frameRate for auto testing
 *  'h'/'H'-key:         hide / show actual digit
 *
 ******************************************************************************/

/******************************************************************************
 *
 *  MOUSE HANDLER
 *
 ******************************************************************************/
public void mousePressed() 
{
  if (help)
  {  
    help = !help;
    return;
  }

  if (mouseButton == LEFT)
  {
    // LEFT MOUSE BUTTON PRESSED: TRAINING
    if (useAutoTesting) frameRate(FRAMERATE_DEFAULT);

    if (useFastTraining)
    { // FAST TRAINING WITHOUT DISPLAYING
      fastTraining(trainingBlocksize);
      nrOfTestedImages   = 0;
      nrOfMatchedImages  = 0;
      networkJustTrained = true;
    } else
    { // SLOW (ANIMATED) TRAINING
      nrOfTestedImages   = 0;
      nrOfMatchedImages  = 0;
      fSuccessPercentage = 0.0f;
      fLastAccuracy      = 0.0f;
      // NUMBER OF IMAGES FOR TRAINING      
      trainingImagesLeft = trainingBlocksize;
    }
  } else
  { 
    // RIGHT MOUSE BUTTON PRESSED: TESTING
    useAutoTesting = !useAutoTesting;
    if (useAutoTesting)
      frameRate(frameRateAutoTesting);
    else
      frameRate(FRAMERATE_DEFAULT);
    // println("useAutoTesting: "+useAutoTesting);
  }
} // mousePressed()


/******************************************************************************
 *
 *  KEYBOARD HANDLER
 *
 ******************************************************************************/
public void keyPressed()
{
  if (help)
  {  
    help = !help;
    return;
  }  

  if (key == ' ')
  {
    // TEST THE NEXT NEW IMAGE
    testingNextImage = true;
  } else if (key == 'f' || key == 'F')
  {
    // TOGGLE BETWEEN FAST- AND SLOW TRAINING
    useFastTraining = !useFastTraining;
    println("useFastTraining: "+useFastTraining);
  } else if (key == 'h' || key == 'H' || key == '?' || key == '/')
  {
    // TOGGLE HELP SCREEN
    help = !help;
    // REDRAW SCREEN
    displayCanvas();
    loop();
  } else if (key == 'l')
  {
    // DECREASE NUMBER OF HIDDEN LAYERS
    if (NR_OF_HIDDEN_LAYERS > 0)
    {  
      NR_OF_HIDDEN_LAYERS--;
      initSettings();
    }
  } else if (key == 'L')
  {
    // INCREASE NUMBER OF HIDDEN LAYERS
    if (NR_OF_HIDDEN_LAYERS < 5)
    {  
      NR_OF_HIDDEN_LAYERS++;
      initSettings();
    }
  } else if (key == 'n')
  {
    // DECREASE NUMBER OF NEURONS IN THE HIDDEN LAYERS
    if (CURRENT_NR_OF_HIDDEN_NEURONS_INDEX > 0)
    {  
      CURRENT_NR_OF_HIDDEN_NEURONS_INDEX--;
      initSettings();
    }
  } else if (key == 'N')
  {
    // DECREASE NUMBER OF NEURONS IN THE HIDDEN LAYERS
    if (CURRENT_NR_OF_HIDDEN_NEURONS_INDEX < POSSIBLE_HIDDEN_NEURONS.length-1)
    {  
      CURRENT_NR_OF_HIDDEN_NEURONS_INDEX++;
      initSettings();
    }
  } else if (key == 't' || key == 'T')
  {
    // TOGGLE LANGUAGE (EN <> NL)
    nl = !nl;
  } else if (key == '+' || key == '=')
  {
    // INCREASE AUTO TESTING SPEED
    frameRateAutoTesting++;
    if (useAutoTesting)
      frameRate(frameRateAutoTesting);
    else
      frameRate(FRAMERATE_DEFAULT);
    println("frameRateAutoTesting: "+frameRateAutoTesting);
  } else if (key == '-' || key == '_')
  {
    // DECREASE AUTO TESTING SPEED
    frameRateAutoTesting--;
    if (useAutoTesting)
      frameRate(frameRateAutoTesting);
    else
      frameRate(FRAMERATE_DEFAULT);
    println("frameRateAutoTesting: "+frameRateAutoTesting);
  } else if (key == 'b')
  {  // DECREASE TRAINING BLOCKSIZE
    if (trainingBlocksize > 100) trainingBlocksize -= 100;
    println("trainingBlocksize: "+trainingBlocksize);
  } else if (key == 'B')
  {  // INCREASE TRAINING BLOCKSIZE
    trainingBlocksize += 100;
    println("trainingBlocksize: "+trainingBlocksize);
  }
} // keyPressed()
/******************************************************************************
 *
 *  LOAD MNIST TRAINING- AND TEST-DATA
 *
 *  http://yann.lecun.com/exdb/mnist/
 *
 ******************************************************************************/

DatasetRecord [] trainingSet;
DatasetRecord [] testingSet;

/******************************************************************************
 *
 *  DATASET RECORD CLASS
 *
 ******************************************************************************/
class DatasetRecord
{
  /****************************************************************************
   *
   *  PROPERTIES
   *
   ****************************************************************************/
  // PIXEL VALUES OF THE IMAGE (-1.0 = WHITE; +1.0 = BLACK)
  float [] inputs;
  // +1.0 INDICATES THE CORRECT ANSWER (THE REST IS -1.0)
  float [] outputs;
  int output;


  /****************************************************************************
   *
   *  CONSTRUCTOR
   *
   ****************************************************************************/
  DatasetRecord()
  {
    inputs  = new float [NR_OF_INPUT_NEURONS];
    outputs = new float [NR_OF_OUTPUT_NEURONS];
  } // DatasetRecord()


  /****************************************************************************
   *
   *  LOAD IMAGES
   *
   ****************************************************************************/
  public void imageLoad(byte [] images, int offset)
  {
    for (int i = 0; i < NR_OF_INPUT_NEURONS; i++)
      inputs[i] = PApplet.parseInt(images[i+offset]) / 128.0f - 1.0f;
  } // imageLoad()


  /****************************************************************************
   *
   *  LOAD LABELS
   *
   *  LABELS ARE THE CORRECT ANSWERS, IN THIS CASE '0'..'9'
   *
   ****************************************************************************/
  public void labelLoad(byte [] labels, int offset)
  {
    output = PApplet.parseInt(labels[offset]);
    for (int i = 0; i < NR_OF_OUTPUT_NEURONS; i++)
      if (i == output)
        outputs[i] = 1.0f;
      else
        outputs[i] = -1.0f;
  } // labelLoad()
} // DatasetRecord


/******************************************************************************
 *
 *  CREATE AND FILL DATASET OBJECTS
 *
 ******************************************************************************/
public void loadData()
{
  // THE IMAGES (196 BYTES PER IMAGE)
  byte [] images = loadBytes("t10k-images-14x14.idx3-ubyte");
  // THE CORRECT OUTPUTS (0..9)
  byte [] labels = loadBytes("t10k-labels.idx1-ubyte");

  // TRAINING IMAGES
  trainingSet = new DatasetRecord [TRAINING_SET_SIZE];
  int trainPos = 0;

  // TESTING IMAGES
  testingSet = new DatasetRecord [TESTING_SET_SIZE];
  int testPos = 0;

  for (int i = 0; i < (TRAINING_SET_SIZE + TESTING_SET_SIZE); i++)
  {
    if (i % 5 != 0)
    {
      // ADD TRAINING DATA RECORD (4/5th OF THE TOTAL AMOUNT OF RECORDS)
      trainingSet[trainPos] = new DatasetRecord();
      trainingSet[trainPos].imageLoad(images, 16 + i * NR_OF_INPUT_NEURONS);
      trainingSet[trainPos].labelLoad(labels, 8 + i);
      trainPos++;
    } else
    {
      // ADD TESTING DATA RECORD (1/5th OF THE TOTAL AMOUNT OF RECORDS)
      testingSet[testPos] = new DatasetRecord();
      testingSet[testPos].imageLoad(images, 16 + i * NR_OF_INPUT_NEURONS);
      testingSet[testPos].labelLoad(labels, 8 + i);
      testPos++;
    }
  }
} // loadData()
/******************************************************************************
 *
 *  NETWORK CLASS
 *
 ******************************************************************************/
class Network
{
  /****************************************************************************
   *
   *  PROPERTIES
   *
   ****************************************************************************/

  // TEMPORARY NEURON ARRAY
  Neuron [] hiddenLayer;

  // INPUT NEURONS
  Neuron [] inputLayerNeurons; 

  // HIDDEN NEURONS (MULTIPLE LAYERS)
  ArrayList<Neuron[]> hiddenLayerNeurons = new ArrayList();

  // OUTPUT NEURONS
  Neuron [] outputLayerNeurons;

  int lastResponse = -1;


  /****************************************************************************
   *
   *  CONSTRUCTOR:
   *
   *  CREATE A NETWORK SPECIFYING NUMBERS OF INPUTS AND OUTPUTS
   *  AND NUMBER OF OUTPUTS, E.G. NETWORK(4,3)
   *
   ****************************************************************************/
  Network(int _inputs, int _outputs)
  {
    // INITIALIZE INPUT NEURONS
    inputLayerNeurons   = new Neuron [_inputs];

    // INITIALIZE HIDDEN NEURONS
    for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++) hiddenLayerNeurons.add(new Neuron [NR_OF_HIDDEN_NEURONS.get(i)]);

    // INITIALIZE OUTPUT NEURONS
    outputLayerNeurons  = new Neuron [_outputs];

    // SET UP THE NETWORK TOPOLOGY
    for (int i = 0; i < inputLayerNeurons.length; i++) inputLayerNeurons[i] = new Neuron();

    if (NR_OF_HIDDEN_LAYERS == 0)
    { 
      // NO HIDDEN LAYERS: INPUT CONNECTS DIRECTLY TO OUTPUT
      for (int k = 0; k < outputLayerNeurons.length; k++) outputLayerNeurons[k] = new Neuron(inputLayerNeurons);
    } else
    {
      // CONNECT THE HIDDEN LAYERS
      for (int m=0; m<NR_OF_HIDDEN_LAYERS; m++)
      {   
        hiddenLayer = hiddenLayerNeurons.get(m);
        if (m == 0)
          // FIRST HIDDEN LAYER: CONNECT TO INPUT LAYER
          for (int j = 0; j < hiddenLayer.length; j++) hiddenLayer[j] = new Neuron(inputLayerNeurons);
        if (m > 0)
          // MIDDLE HIDDEN LAYER: CONNECT TO THE PREVIOUS HIDDEN LAYER
          for (int j = 0; j < hiddenLayer.length; j++) hiddenLayer[j] = new Neuron(hiddenLayerNeurons.get(m-1));
        if (m == NR_OF_HIDDEN_LAYERS-1)
          // LAST HIDDEN LAYER: CONNECT TO OUTPUT LAYER
          for (int k = 0; k < outputLayerNeurons.length; k++)
            outputLayerNeurons[k] = new Neuron(hiddenLayer);
        // UPDATE LAYER
        hiddenLayerNeurons.set(m, hiddenLayer);
      } // for (int m=0; m<NR_OF_HIDDEN_LAYERS; m++)
    } // if (NR_OF_HIDDEN_LAYERS == 0)
  } // Network


  /****************************************************************************
   *
   *  CALCULATE THE OUTPUTS FOR A TESTING OR TRAINING IMAGE (inputs)
   *
   ****************************************************************************/
  public int respond(float [] inputs)
  {
    float [] responses = new float [outputLayerNeurons.length];

    // FEED FORWARD
    // SIMPLY SET THE INPUT LAYER TO DISPLAY THE INPUTS
    for (int i = 0; i < inputLayerNeurons.length; i++) inputLayerNeurons[i].neuronOutput = inputs[i];

    // FEED FORWARD THE HIDDEN LAYER(S)
    for (int j = 0; j < NR_OF_HIDDEN_LAYERS; j++)
    {
      hiddenLayer = hiddenLayerNeurons.get(j);
      for (int n = 0; n < hiddenLayer.length; n++) hiddenLayer[n].respond();
    }

    // AND FINALLY FEED FORWARD TO THE OUTPUT LAYER
    for (int k = 0; k < outputLayerNeurons.length; k++) responses[k] = outputLayerNeurons[k].respond();

    // NOW CHECK THE BEST RESPONSE:
    int response = -1;
    float best = max(responses);
    for (int a = 0; a < responses.length; a++)
      if (responses[a] == best) response = a;

    fLastAccuracy = (best+1) * 50;

    lastResponse = response;

    return response;
  } // respond()


  /****************************************************************************
   *
   *  TRAIN THE NETWORK
   *
   ****************************************************************************/
  public void train(float [] outputs)
  {
    // ADJUST THE OUTPUT LAYER
    for (int k = 0; k < outputLayerNeurons.length; k++)
    {
      outputLayerNeurons[k].finderror(outputs[k]);
      outputLayerNeurons[k].train();
    }

    // PROPAGATE BACK TO HIDDEN LAYERS
    for (int j = 0; j < NR_OF_HIDDEN_LAYERS; j++)
    {
      hiddenLayer = hiddenLayerNeurons.get(j);
      for (int n = 0; n < hiddenLayer.length; n++) hiddenLayer[n].train();
    }

    // THE INPUT LAYER DOESN'T LEARN:
    // IT IS SIMPLY THE INPUTS
  } // train()


  /****************************************************************************
   *
   *  DRAW THE NETWORK
   *
   ****************************************************************************/
  public void draw()
  {
    // NOTE, THIS DRAW IS HARD-CODED FOR NETWORK(196,49,10)
    // WHICH REFLECTS MY USE OF THE MNIST DATABASE OF HANDWRITTEN DIGITS

    pushMatrix();

    // OUTPUT LAYER
    // THIS IS SLIGHTLY TRICKY -- I'VE SWITCHED THE ORDER SO THE OUTPUT
    // NEURONS ARE ARRANGE 1,2,3...8,9,0 RATHER THAN 0,1,2...7,8,9
    // (THAT'S WHAT THE (K+9) % 10 IS DOING)
    for (int k = 0; k < outputLayerNeurons.length; k++) {
      pushMatrix();
      translate(((k+9)%10) * 600 / 20.0f + output_offset_x, 20);
      outputLayerNeurons[k].draw("o", k);
      popMatrix();
    }

    // HIDDEN LAYER(S)
    for (int l=0; l<NR_OF_HIDDEN_LAYERS; l++)
    {
      hiddenLayer = hiddenLayerNeurons.get(l);  
      for (int m = 0; m < hiddenLayer.length; m++)
      {
        pushMatrix();
        translate((m % HIDDEN_COLUMNS[l]) * 600 / 30.0f + hidden_offset_x[l], (m / HIDDEN_COLUMNS[l]) * 600 / 30.0f + hidden_offset_y[l]);
        hiddenLayer[m].draw("h", m);
        popMatrix();
      }
    }

    // INPUT LAYER
    for (int i = 0; i < inputLayerNeurons.length; i++)
    {
      pushMatrix();
      translate((i%INPUT_COLUMNS) * 600 / 30.0f + input_offset_x, (i/INPUT_COLUMNS) * 600 / 30.0f + input_offset_y);
      inputLayerNeurons[i].draw("i", i);
      popMatrix();
    }

    popMatrix();
  } // draw()
} // Network
/******************************************************************************
 *
 *  NEURON CLASS
 *
 ******************************************************************************/
float LEARNING_RATE = 0.01f;

class Neuron
{
  Neuron [] inputNeurons;
  float [] neuronWeights;
  float neuronTreshold = 0.0f;
  float neuronOutput   = 0.0f;
  float neuronError;


  /****************************************************************************
   *
   *  CONSTRUCTOR FOR THE INPUT LAYER
   *
   ****************************************************************************/
  Neuron()
  {
    // INITIAL RANDOM OUTPUT
    neuronOutput = sigmoid.lookupSigmoid(random(-5.0f, 5.0f));
  } // Neuron()


  /****************************************************************************
   *
   *  CONSTRUCTOR FOR THE HIDDEN- AND OUTPUT LAYER
   *
   ****************************************************************************/
  Neuron(Neuron[] inputs)
  {
    inputNeurons  = new Neuron [inputs.length];
    neuronWeights = new float [inputs.length];

    // CREATE RANDOM WEIGHTS
    for (int i = 0; i < inputs.length; i++)
    {
      inputNeurons[i] = inputs[i];
      neuronWeights[i] = random(-1.0f, 1.0f);
    }

    // RANDOM TRESHOLD
    neuronTreshold = random(-1.0f, 1.0f);
    neuronError = 0.0f;

    // INITIAL RANDOM OUTPUT
    neuronOutput = sigmoid.lookupSigmoid(random(-5.0f, 5.0f));
  } // Neuron()


  /****************************************************************************
   *
   *  RESPOND LOOKS AT THE LAYER BELOW, AND PREPARES A RESPONSE
   *
   ****************************************************************************/
  public float respond()
  {
    float input = 0.0f;
    for (int i = 0; i < inputNeurons.length; i++) {
      input += inputNeurons[i].neuronOutput * neuronWeights[i];
    }
    neuronOutput = sigmoid.lookupSigmoid(input + neuronTreshold);
    // RESET OUR ERROR VALUE READY FOR TRAINING
    neuronError = 0.0f;

    return neuronOutput;
  } // respond()


  /****************************************************************************
   *
   *  FIND ERROR IS USED ON THE OUTPUT NEURONS
   *
   ****************************************************************************/
  public void finderror(float desired)
  {
    neuronError = desired - neuronOutput;
  } // finderror()


  /****************************************************************************
   *
   *  TRAIN ADJUSTS THE WEIGHTS BY COMPARING ACTUAL OUTPUT TO CORRECT OUTPUT
   *
   ****************************************************************************/
  public void train()
  {
    float delta = (1.0f - neuronOutput) * (1.0f + neuronOutput) * neuronError * LEARNING_RATE;
    for (int i = 0; i < inputNeurons.length; i++)
    {
      // TELL THE NEXT LAYER DOWN WHAT IT'S DOING WRONG
      inputNeurons[i].neuronError += neuronWeights[i] * neuronError;
      // CORRECT OUR WEIGHTS
      neuronWeights[i] += inputNeurons[i].neuronOutput * delta;
    }
  } // train()


  /****************************************************************************
   *
   *  DRAW THE NEURONS FOR ALL LAYERS
   *
   *  not_hidden: in- or output layer
   *  type: "i", "h" or "o" (input, hidden, output)
   *  nr: sequence number of the neuron
   *
   ****************************************************************************/
  public void draw(String type, int nr)
  {
    float level = (0.5f-neuronOutput*0.5f);

    if (type != "h")
    { // INPUT- OR OUTPUT LAYER
      fill(255 * level);
      if (type == "o")
      { // OUPUT LAYER
        // PRINTS THE NUMBERS ABOUT THE OUTPUT DOTS
        pushStyle();
        fill(0);
        textSize(12);
        text(nr, 15, -6);
        popStyle();
      }
    } else {
      // MERGE HIDDEN LAYER WITH BACKGROUND COLOR
      fill(110 + 128 * level, 102 + 128 * level, 127 + 128 * level);
    }

    // DRAW THE DOT
    if (type == "i")
      // INPUT
      rect(0, 0, INPUT_SQUARE_SIZE, INPUT_SQUARE_SIZE);
    else if (type  == "h")
      // HIDDEN
      ellipse(0, 0, HIDDEN_DOT_SIZE, HIDDEN_DOT_SIZE);
    else
      // OUTPUT
      ellipse(0, 0, OUTPUT_DOT_SIZE, OUTPUT_DOT_SIZE);
  } // draw()
} // Neuron
/******************************************************************************
 *
 *  GLOBAL SETTINGS AND VARIABLES
 *
 ******************************************************************************/

// LANGUAGE
boolean nl   = false;

// HELP SCREEN
boolean help = false;
PImage  numberImg;
boolean running = false;

// NUMBER OF IMAGES TO PROCESS DURING ONE TRAINING SESSION
final static int TRAINING_BLOCKSIZE = 1000;

/******************************************************************************
 *  INPUT LAYER
 ******************************************************************************/
// EACH IMAGE IS 14 X 14 PIXELS = 196 PIXELS IN TOTAL
final static int NR_OF_INPUT_NEURONS = 14 * 14;
// NUMBER OF DISPLAY COLUMNS
final static int INPUT_COLUMNS  = PApplet.parseInt(sqrt(NR_OF_INPUT_NEURONS));
// SIZE IN PIXELS
final static int INPUT_SIZE = 280;
final static int INPUT_SQUARE_SIZE = 18;

/******************************************************************************
 *  HIDDEN LAYER(S)
 ******************************************************************************/
int   NR_OF_HIDDEN_LAYERS  = 1;
int[] POSSIBLE_HIDDEN_NEURONS = {
  5*5, 6*6, 7*7, 8*8, 9*9, 10*10, 11*11, 12*12
};
int   CURRENT_NR_OF_HIDDEN_NEURONS_INDEX = 2;
ArrayList<Integer> NR_OF_HIDDEN_NEURONS = new ArrayList();

// NUMBER OF COLUMNS IN THE HIDDEN LAYER
int[] HIDDEN_COLUMNS  = new int[NR_OF_HIDDEN_LAYERS];
// SIZE OF THE HIDDEN LAYERS (IN PIXELS)
int[] HIDDEN_SIZES    = new int[NR_OF_HIDDEN_LAYERS];
final static int   HIDDEN_DOT_SIZE = 16;

/******************************************************************************
 *  OUTPUT LAYER
 ******************************************************************************/
// DIGITS '0'..'9'
final static int NR_OF_OUTPUT_NEURONS  = 10;
// HEIGHT OF THE OUTPUT LAYER (IN PIXELS)
final static int OUTPUT_HEIGHT   = 43;
// WIDTH OF THE OUTPUT LAYER (IN PIXELS)
final static int OUTPUT_WIDTH    = 293;
final static int OUTPUT_DOT_SIZE = 22;

/******************************************************************************
 *  TRAINING AND TESTING
 ******************************************************************************/
final static int TRAINING_SET_SIZE = 8000;
final static int TESTING_SET_SIZE  = 2000;

// BACKGROUND COLOR
int BACKGROUND_COLOR = color(220, 204, 255); // LIGHT PURPLE
// color BACKGROUND_COLOR = color(221, 237, 189);    // LIGHT GREEN

// FRAMERATE FOR THE AUTO TESTING MODE
final static int FRAMERATE_AUTO_TESTING = 25;
final static int FRAMERATE_DEFAULT      = 200;

// THE AMOUNT OF IMAGES TO TEST BEFORE PRINTING THE RESULT AND STOP (IN AUTO TEST MODE)
final static int CHECKPOINT_AUTO_TESTING = 1000;

// LEFT CLICK TO TEST, RIGHT CLICK (OR CTRL+CLICK ON A MAC) TO TRAIN
boolean testingNextImage   = true;
boolean networkJustTrained = false;

// USE THE FAST TRAINING (WITHOUT DISPLAY) OR THE SLOW TRAINING (WITH ANIMATION)
boolean useFastTraining    = false;

// AUTOMATICALLY FEED TESTING IMAGES
boolean useAutoTesting     = false;
int frameRateAutoTesting   = FRAMERATE_AUTO_TESTING;

// NUMBER OF IMAGES TO USE DURING ONE TRAINING SESSION
int trainingBlocksize      = TRAINING_BLOCKSIZE;

boolean useFullScreen = false;

// SHOW / HIDE ACTUAL DIGIT
boolean hideActual         = false;

// COUNTERS
int nrOfTestedImages    = 0;
int nrOfMatchedImages   = 0;
int nrOfTrainedImages   = 0;

// NUMBER OF IMAGES LEFT FOR THE CURRENT TRAINING SESSION
int trainingImagesLeft  = 0;

float fSuccessPercentage = 0.0f;

int response = -1;
int actual   = -1;

// ACCURACY OF THE LAST TESTED IMAGE
float fLastAccuracy = 0.0f;

// LAYOUT
int input_offset_x;
int input_offset_y;

int[] hidden_offset_x = new int[NR_OF_HIDDEN_LAYERS];
int[] hidden_offset_y = new int[NR_OF_HIDDEN_LAYERS];

int output_offset_x;
int output_offset_y;

// VECTORS FOR THE CONNECTORS
PVector   inputCenterTop;
PVector[] hiddenCenterBottom = new PVector[NR_OF_HIDDEN_LAYERS];
PVector[] hiddenCenterTop    = new PVector[NR_OF_HIDDEN_LAYERS];
PVector[] hiddenCenterLeft   = new PVector[NR_OF_HIDDEN_LAYERS];
PVector[] hiddenCenterRight  = new PVector[NR_OF_HIDDEN_LAYERS];

PVector   outputCenterBottom;
PVector   outputLeftBottom;

// ALTERNATING COLORS FOR THE CONNECTORS
int     color1 = color(243, 238, 254);
int     color2 = color(181, 162, 222);
int     currentColor = color1;


/******************************************************************************
 *
 *  SETUP
 *
 ******************************************************************************/
public void initSettings()
{
  numberImg = loadImage("number.png");

  // NUMBER OF COLUMNS IN THE HIDDEN GRID
  HIDDEN_COLUMNS  = new int[NR_OF_HIDDEN_LAYERS];
  // SIZE OF THE HIDDEN LAYERS (IN PIXELS)
  HIDDEN_SIZES    = new int[NR_OF_HIDDEN_LAYERS];  

  // LAYOUT
  hidden_offset_x = new int[NR_OF_HIDDEN_LAYERS];
  hidden_offset_y = new int[NR_OF_HIDDEN_LAYERS];  

  // VECTORS FOR THE CONNECTORS
  hiddenCenterBottom = new PVector[NR_OF_HIDDEN_LAYERS];
  hiddenCenterTop    = new PVector[NR_OF_HIDDEN_LAYERS];
  hiddenCenterLeft   = new PVector[NR_OF_HIDDEN_LAYERS];
  hiddenCenterRight  = new PVector[NR_OF_HIDDEN_LAYERS];

  NR_OF_HIDDEN_NEURONS = new ArrayList();

  for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
  {
    NR_OF_HIDDEN_NEURONS.add(POSSIBLE_HIDDEN_NEURONS[CURRENT_NR_OF_HIDDEN_NEURONS_INDEX]);
    // NUMBER OF COLUMNS IN THE HIDDEN LAYER GRID
    HIDDEN_COLUMNS [i] = PApplet.parseInt(sqrt(NR_OF_HIDDEN_NEURONS.get(i)));
    // SIZE OF THE HIDDEN LAYER (IN PIXELS)
    HIDDEN_SIZES   [i] = PApplet.parseInt(HIDDEN_COLUMNS[i]*HIDDEN_DOT_SIZE+(HIDDEN_COLUMNS[i]-1)*4);
  }

  // CALCULATE POSITIONS FOR DISPLAY
  input_offset_x  = (width-INPUT_SIZE)/2;
  input_offset_y  = height-INPUT_SIZE-20;

  output_offset_x = (width-OUTPUT_WIDTH)/2;
  output_offset_y = 42;

  // CONNECTIONS
  inputCenterTop     = new PVector(input_offset_x+(INPUT_SIZE>>1), input_offset_y);
  outputCenterBottom = new PVector(output_offset_x+(OUTPUT_WIDTH>>1), output_offset_y);
  outputLeftBottom   = new PVector(output_offset_x, output_offset_y);

  int hSpace = 0;
  int vSpace = 0;

  // HIDDEN LAYERS
  if (NR_OF_HIDDEN_LAYERS > 0)
  {
    hSpace = (width - (NR_OF_HIDDEN_LAYERS * HIDDEN_SIZES[0])) / (NR_OF_HIDDEN_LAYERS + 1);
    vSpace = (height - OUTPUT_HEIGHT - HIDDEN_SIZES[0] - INPUT_SIZE - 20) / 2;
    for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
    { 
      hidden_offset_x[i]    = i * HIDDEN_SIZES[0] + (i + 1) * hSpace;
      hidden_offset_y[i]    = vSpace + OUTPUT_HEIGHT;
      hiddenCenterTop[i]    = new PVector(hidden_offset_x[i] + (HIDDEN_SIZES[0]>>1), hidden_offset_y[i]);      
      hiddenCenterBottom[i] = new PVector(hidden_offset_x[i] + (HIDDEN_SIZES[0]>>1), hidden_offset_y[i] + HIDDEN_SIZES[0]);
      hiddenCenterLeft[i]   = new PVector(hidden_offset_x[i], hidden_offset_y[i] + (HIDDEN_SIZES[0]>>1));
      hiddenCenterRight[i]  = new PVector(hidden_offset_x[i] + HIDDEN_SIZES[0], hidden_offset_y[i] + (HIDDEN_SIZES[0]>>1));
    }
  }

  // LOAD SIGMOID
  sigmoid = new Sigmoid();   

  // CREATE THE NETWORK
  neuralnet = new Network(NR_OF_INPUT_NEURONS, NR_OF_OUTPUT_NEURONS);  

  nrOfTrainedImages  = 0;
  nrOfTestedImages   = 0;
  nrOfMatchedImages  = 0;
  fSuccessPercentage = 0.0f;
  fLastAccuracy      = 0.0f;

  // FORCE THE DISPLAY TO REDRAW
  testingNextImage = true;
} // initSettings()
/******************************************************************************
 *
 *  SIGMOID FUNCTION
 *
 *  A SIGMOID FUNCTION IS THE NEURON'S RESPONSE TO INPUTS
 *  THE SIGMOIDAL RESPONSE RANGES FROM -1.0 TO 1.0
 *
 *  FOR EXAMPLE, THE WEIGHTED SUM OF INPUTS MIGHT BE "2.1"
 *  OUR RESPONSE WOULD BE LOOKUPSIGMOID(2.1) = 0.970
 *
 ******************************************************************************/
class Sigmoid
{
  // THIS IS A LOOK UP TABLE FOR SIGMOID (NEURAL RESPONSE) VALUES
  // WHICH IS VALID FROM -5.0 TO 5.0 AND WILL RETURN FROM -1.0 TO 1.0
  float [] g_sigmoid = new float [200];  


  /****************************************************************************
   *
   *  CONSTRUCTOR
   *
   ****************************************************************************/
  Sigmoid()
  {
    setupSigmoid();
  } // Sigmoid()


  /****************************************************************************
   *
   *  THIS FUNCTION PRECALCULATE A SIGMOID (RESPONSE) FUNCTION
   *
   ****************************************************************************/
  public void setupSigmoid()
  {
    for (int i = 0; i < 200; i++)
    {
      float x = (i / 20.0f) - 5.0f;
      g_sigmoid[i] = 2.0f / (1.0f + exp(-2.0f * x)) - 1.0f;
    }
  } // setupSigmoid()


  /****************************************************************************
   *
   *  ONCE THE SIGMOID HAS BEEN SET UP, THIS FUNCTION ACCESSES IT
   *
   ****************************************************************************/
  public float lookupSigmoid(float x)
  {
    return g_sigmoid[constrain((int) floor((x + 5.0f) * 20.0f), 0, 199)];
  } // lookupSigmoid()
}
  public void settings() {  size(1024, 800); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "--present", "--window-color=#000000", "--hide-stop", "rvg_neural_net20" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
