/******************************************************************************
 *
 *  A SIMPLE NEURAL NETWORK v2.0
 *
 *  v12/05/2015  (c) Rolf van Gelder, Eindhoven, NL - http://cagewebdev.com
 *
 ******************************************************************************/

// THE NETWORK
Network neuralnet;

// SIGMOID LOOKUP TABLE
Sigmoid sigmoid;


/******************************************************************************
 *
 *  SETUP
 *
 ******************************************************************************/
void setup()
{
  size(1024, 800);

  // INITIALIZE THE SETTINGS
  initSettings();

  // LOAD TRAINING- AND TEST-DATASETS
  loadData();

  // LAYOUT DEFAULTS
  background(BACKGROUND_COLOR);
  noStroke();
  ellipseMode(CORNER);
  textSize(15);
  textAlign(RIGHT);

  if (useAutoTesting)
    frameRate(frameRateAutoTesting);
  else
    frameRate(FRAMERATE_DEFAULT);
} // setup()


/******************************************************************************
 *
 *  DRAW
 *
 ******************************************************************************/
void draw()
{
  if (help)
  {  
    helpScreen();
  } else
  {
    if (trainingImagesLeft > 0)
    { // TRAINING SESSION IS RUNNING

      running = true;

      // GET NEXT TRAINING IMAGE
      trainOneImage();

      // ONE LESS TO TRAIN...
      trainingImagesLeft--;

      if (trainingImagesLeft < 1)
      { // LAST IMAGE PROCESSED: RESET COUNTERS
        nrOfTestedImages  = 0;
        nrOfMatchedImages = 0;
        if (useAutoTesting) frameRate(frameRateAutoTesting);
      }
    } else
    {
      running = false;
    } // if (trainingImagesLeft > 0)

    // TEST THE NEXT IMAGE
    if (useAutoTesting) testingNextImage = true;

    if (testingNextImage || networkJustTrained)
    { 
      if (testingNextImage)
      { 
        running = true;
        // TEST A RANDOM TEST IMAGE
        int row = (int) floor(random(0, testingSet.length));
        response = neuralnet.respond(testingSet[row].inputs);
        actual   = testingSet[row].output;
      }
    } else
      running = false;
    // if (testingNextImage || networkJustTrained)

    // DRAW THE NETWORK LAYERS AND STATISTICS
    displayCanvas();
    loop();
  } // if (help)
} // draw()


/******************************************************************************
 *
 *  DRAW SOME STATISTICS
 *
 ******************************************************************************/
void drawStats()
{
  int margin  = 10;
  int spacing = 24;
  int top     = height-INPUT_SIZE-5;

  // DRAW CONNECTIONS BETWEEN THE LAYERS
  drawConnections();

  // DRAW SOME STATISTICS
  fill(0);
  textAlign(LEFT);
  if (trainingImagesLeft > 0)
  {
    fill(255);
    if (nl)
      rect(margin - 3, 18, 145, 19);
    else
      rect(margin - 3, 18, 155, 19);

    fill(0);
    if (nl)
      text("NETWERK TRAINEN", margin, 33);
    else
      text("TRAINING NETWORK", margin, 33);
  } else if (testingNextImage)
  {
    fill(255);
    if (nl)
      rect(margin - 3, 18, 135, 19);
    else
      rect(margin - 3, 18, 145, 19);

    fill(0);
    if (nl)
      text("NETWERK TESTEN", margin, 33);
    else
      text("TESTING NETWORK", margin, 33);    
    if (response == actual) nrOfMatchedImages++;
    nrOfTestedImages++;
  }
  if (nl)
  {
    text("getrainde afbeeldingen: "+nrOfTrainedImages, margin, top);
    text("geteste afbeeldingen: "+nrOfTestedImages, margin, top+spacing);
    text("treffers: "+nrOfMatchedImages, margin, top+2*spacing);
  } else
  {  
    text("trained images: "+nrOfTrainedImages, margin, top);
    text("tested images: "+nrOfTestedImages, margin, top+spacing);
    text("hits: "+nrOfMatchedImages, margin, top+2*spacing);
  }
  fSuccessPercentage = 0.0;
  if (nrOfTestedImages > 0) fSuccessPercentage = (float)nrOfMatchedImages*100/nrOfTestedImages;
  if ((nrOfTestedImages%CHECKPOINT_AUTO_TESTING) == 0 && useAutoTesting)
  {
    if (nl)
      println("getrainde afbeeldingen: " +nrOfTrainedImages+" geteste afbeeldingen: "+CHECKPOINT_AUTO_TESTING+" succes: "+fSuccessPercentage+"%");
    else
      println("trained images: " +nrOfTrainedImages+" tested images: "+CHECKPOINT_AUTO_TESTING+" success: "+fSuccessPercentage+"%");
    useAutoTesting = false;
  }
  if (nl)
  {
    text("gemiddeld succes: "+nfc(fSuccessPercentage, 1)+"%", margin, top+3*spacing);
    text("betrouwbaarheid (laatste test): "+nfc(fLastAccuracy, 1)+"%", margin, top+4*spacing);
  } else
  {
    text("average success: "+nfc(fSuccessPercentage, 1)+"%", margin, top+3*spacing);
    text("accuracy (last test): "+nfc(fLastAccuracy, 1)+"%", margin, top+4*spacing);
  }
  pushStyle();
  textAlign(CENTER);
  String layers = "";
  if (NR_OF_HIDDEN_LAYERS > 0)
  {  
    if (NR_OF_HIDDEN_LAYERS > 1)
    {  
      if (nl)
        layers = "verborgen lagen";
      else
        layers = "hidden layer";
    } else
    {  
      if (nl)
        layers = "verborgen laag";
      else
        layers = "hidden layer";
    }
  }
  for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
  {
    if (nl)
      text(layers+" ("+NR_OF_HIDDEN_NEURONS.get(i)+" neuronen)", width>>1, hidden_offset_y[i] - 10);
    else
      text(layers+" ("+NR_OF_HIDDEN_NEURONS.get(i)+" neurons)", width>>1, hidden_offset_y[i] - 10);
  }
  popStyle();

  textAlign(LEFT);
  if (nl)
    text("type 'h' voor hulp", margin, height-25);
  else
    text("press 'h' for help", margin, height-25);

  textAlign(RIGHT);
  text("fps: "+nfc(frameRate, 2), width-margin, height-25);

  // ACTUAL AND RESPONSE
  if (trainingImagesLeft < 1 && nrOfTestedImages > 1)
  {
    color bgcolor, fgcolor;
    if (response != actual)
    {  
      bgcolor = color(255, 0, 0);
      fgcolor = color(255);
    } else
    {  
      bgcolor = color(0, 255, 0);
      fgcolor = color(0);
    }
    fill(bgcolor);
    if (nl)
      rect(width-margin-94, 20, 99, 20);
    else
      rect(width-margin-90, 20, 95, 20);
    fill(fgcolor);
    if (nl)
      text("antwoord: "+str(response), width-margin, 35);
    else
      text("response: "+str(response), width-margin, 35);
    if (!hideActual)
    {
      fill(bgcolor);
      if (nl)
        rect(width-margin-88, input_offset_y-2, 92, 20);
      else
        rect(width-margin-68, input_offset_y-2, 72, 20);
      fill(fgcolor);
      if (nl)
        text("werkelijk: "+str(actual), width-margin, input_offset_y+12);
      else
        text("actual: "+str(actual), width-margin, input_offset_y+12);
    }
  }

  // STOP THE DISPLAY LOOP
  testingNextImage   = false;
  networkJustTrained = false;
} // drawStats()


/******************************************************************************
 *
 *  DRAW CONNECTIONS BETWEEN THE LAYERS
 *
 ******************************************************************************/
void drawConnections()
{
  // LET THE OUTPUT CONNECTOR POINT TO THE STRONGEST VALUE
  // neuralnet.lastResponse: 0..9 => 1..10
  int r = neuralnet.lastResponse;
  if (r == 0) r = 10;
  int outputX = int(outputLeftBottom.x + OUTPUT_DOT_SIZE/2 + (r-1)*(OUTPUT_DOT_SIZE+8));

  pushStyle();

  // TOGGLE THE COLOR OF THE CONNECTION
  stroke(currentColor);
  if (running)
  {
    if (currentColor == color1)
      currentColor = color2;
    else
      currentColor = color1;
  }

  strokeWeight(3);

  if (NR_OF_HIDDEN_LAYERS == 0)
  {
    // CONNECT INPUT DIRECTLY WITH OUTPUT
    line(inputCenterTop.x, inputCenterTop.y, outputCenterBottom.x, outputCenterBottom.y);
  } else
  {
    // CONNECT HIDDEN LAYERS
    for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
    {  
      if (i == 0)
        // FIRST HIDDEN LAYER: CONNECT TO INPUTS
        line(inputCenterTop.x, inputCenterTop.y, hiddenCenterBottom[0].x, hiddenCenterBottom[0].y);
      if (i > 0)
        // MIDDLE HIDDEN LAYER: CONNECT TO PREVIOUS HIDDEN LAYER
        line(hiddenCenterLeft[i].x, hiddenCenterLeft[i].y, hiddenCenterRight[i-1].x, hiddenCenterRight[i-1].y);
      if (i == NR_OF_HIDDEN_LAYERS-1)
        // LAST HIDDEN LAYER: CONNECT TO OUTPUTS
        line(hiddenCenterTop[i].x, hiddenCenterTop[i].y, outputX, outputLeftBottom.y-OUTPUT_DOT_SIZE/2);
    } // for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
  } // if (NR_OF_HIDDEN_LAYERS == 0)

  popStyle();
} // drawConnections()


/******************************************************************************
 *
 *  DRAW THE NETWORK LAYERS AND STATISTICS
 *
 ******************************************************************************/
void displayCanvas()
{
  background(BACKGROUND_COLOR);

  // CALCULATE AND DRAW THE NETWORK LAYERS 
  neuralnet.draw();

  // DRAW SOME STATISTICS
  drawStats();
} // displayCanvas()


/******************************************************************************
 *
 *  FAST TRAINING WITHOUT DISPLAYING
 *
 ******************************************************************************/
void fastTraining(int cnt)
{
  // TRAIN THE NETWORK WITH A SET OF IMAGES
  for (int i = 0; i < cnt; i++) trainOneImage();
} // fastTraining()


/******************************************************************************
 *
 *  TRAIN THE NETWORK WITH ONE RANDOM IMAGE
 *
 ******************************************************************************/
void trainOneImage()
{
  // SELECT A RANDOM TRAINING INPUT AND TRAIN
  int i = (int)floor(random(0, trainingSet.length));
  // WHAT THE NETWORK THINKS IT IS
  response = neuralnet.respond(trainingSet[i].inputs);
  // WHAT IT ACTUALLY IS
  actual   = trainingSet[i].output;
  neuralnet.train(trainingSet[i].outputs);
  nrOfTrainedImages++;
} // trainOneImage()