/******************************************************************************
 *
 *  GLOBAL SETTINGS AND VARIABLES
 *
 ******************************************************************************/

// LANGUAGE
boolean nl   = false;

// HELP SCREEN
boolean help = false;
PImage  numberImg;
boolean running = false;

// NUMBER OF IMAGES TO PROCESS DURING ONE TRAINING SESSION
final static int TRAINING_BLOCKSIZE = 1000;

/******************************************************************************
 *  INPUT LAYER
 ******************************************************************************/
// EACH IMAGE IS 14 X 14 PIXELS = 196 PIXELS IN TOTAL
final static int NR_OF_INPUT_NEURONS = 14 * 14;
// NUMBER OF DISPLAY COLUMNS
final static int INPUT_COLUMNS  = int(sqrt(NR_OF_INPUT_NEURONS));
// SIZE IN PIXELS
final static int INPUT_SIZE = 280;
final static int INPUT_SQUARE_SIZE = 18;

/******************************************************************************
 *  HIDDEN LAYER(S)
 ******************************************************************************/
int   NR_OF_HIDDEN_LAYERS  = 1;
int[] POSSIBLE_HIDDEN_NEURONS = {
  5*5, 6*6, 7*7, 8*8, 9*9, 10*10, 11*11, 12*12
};
int   CURRENT_NR_OF_HIDDEN_NEURONS_INDEX = 2;
ArrayList<Integer> NR_OF_HIDDEN_NEURONS = new ArrayList();

// NUMBER OF COLUMNS IN THE HIDDEN LAYER
int[] HIDDEN_COLUMNS  = new int[NR_OF_HIDDEN_LAYERS];
// SIZE OF THE HIDDEN LAYERS (IN PIXELS)
int[] HIDDEN_SIZES    = new int[NR_OF_HIDDEN_LAYERS];
final static int   HIDDEN_DOT_SIZE = 16;

/******************************************************************************
 *  OUTPUT LAYER
 ******************************************************************************/
// DIGITS '0'..'9'
final static int NR_OF_OUTPUT_NEURONS  = 10;
// HEIGHT OF THE OUTPUT LAYER (IN PIXELS)
final static int OUTPUT_HEIGHT   = 43;
// WIDTH OF THE OUTPUT LAYER (IN PIXELS)
final static int OUTPUT_WIDTH    = 293;
final static int OUTPUT_DOT_SIZE = 22;

/******************************************************************************
 *  TRAINING AND TESTING
 ******************************************************************************/
final static int TRAINING_SET_SIZE = 8000;
final static int TESTING_SET_SIZE  = 2000;

// BACKGROUND COLOR
color BACKGROUND_COLOR = color(220, 204, 255); // LIGHT PURPLE
// color BACKGROUND_COLOR = color(221, 237, 189);    // LIGHT GREEN

// FRAMERATE FOR THE AUTO TESTING MODE
final static int FRAMERATE_AUTO_TESTING = 25;
final static int FRAMERATE_DEFAULT      = 200;

// THE AMOUNT OF IMAGES TO TEST BEFORE PRINTING THE RESULT AND STOP (IN AUTO TEST MODE)
final static int CHECKPOINT_AUTO_TESTING = 1000;

// LEFT CLICK TO TEST, RIGHT CLICK (OR CTRL+CLICK ON A MAC) TO TRAIN
boolean testingNextImage   = true;
boolean networkJustTrained = false;

// USE THE FAST TRAINING (WITHOUT DISPLAY) OR THE SLOW TRAINING (WITH ANIMATION)
boolean useFastTraining    = false;

// AUTOMATICALLY FEED TESTING IMAGES
boolean useAutoTesting     = false;
int frameRateAutoTesting   = FRAMERATE_AUTO_TESTING;

// NUMBER OF IMAGES TO USE DURING ONE TRAINING SESSION
int trainingBlocksize      = TRAINING_BLOCKSIZE;

boolean useFullScreen = false;

// SHOW / HIDE ACTUAL DIGIT
boolean hideActual         = false;

// COUNTERS
int nrOfTestedImages    = 0;
int nrOfMatchedImages   = 0;
int nrOfTrainedImages   = 0;

// NUMBER OF IMAGES LEFT FOR THE CURRENT TRAINING SESSION
int trainingImagesLeft  = 0;

float fSuccessPercentage = 0.0;

int response = -1;
int actual   = -1;

// ACCURACY OF THE LAST TESTED IMAGE
float fLastAccuracy = 0.0;

// LAYOUT
int input_offset_x;
int input_offset_y;

int[] hidden_offset_x = new int[NR_OF_HIDDEN_LAYERS];
int[] hidden_offset_y = new int[NR_OF_HIDDEN_LAYERS];

int output_offset_x;
int output_offset_y;

// VECTORS FOR THE CONNECTORS
PVector   inputCenterTop;
PVector[] hiddenCenterBottom = new PVector[NR_OF_HIDDEN_LAYERS];
PVector[] hiddenCenterTop    = new PVector[NR_OF_HIDDEN_LAYERS];
PVector[] hiddenCenterLeft   = new PVector[NR_OF_HIDDEN_LAYERS];
PVector[] hiddenCenterRight  = new PVector[NR_OF_HIDDEN_LAYERS];

PVector   outputCenterBottom;
PVector   outputLeftBottom;

// ALTERNATING COLORS FOR THE CONNECTORS
color     color1 = color(243, 238, 254);
color     color2 = color(181, 162, 222);
color     currentColor = color1;


/******************************************************************************
 *
 *  SETUP
 *
 ******************************************************************************/
void initSettings()
{
  numberImg = loadImage("number.png");

  // NUMBER OF COLUMNS IN THE HIDDEN GRID
  HIDDEN_COLUMNS  = new int[NR_OF_HIDDEN_LAYERS];
  // SIZE OF THE HIDDEN LAYERS (IN PIXELS)
  HIDDEN_SIZES    = new int[NR_OF_HIDDEN_LAYERS];  

  // LAYOUT
  hidden_offset_x = new int[NR_OF_HIDDEN_LAYERS];
  hidden_offset_y = new int[NR_OF_HIDDEN_LAYERS];  

  // VECTORS FOR THE CONNECTORS
  hiddenCenterBottom = new PVector[NR_OF_HIDDEN_LAYERS];
  hiddenCenterTop    = new PVector[NR_OF_HIDDEN_LAYERS];
  hiddenCenterLeft   = new PVector[NR_OF_HIDDEN_LAYERS];
  hiddenCenterRight  = new PVector[NR_OF_HIDDEN_LAYERS];

  NR_OF_HIDDEN_NEURONS = new ArrayList();

  for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
  {
    NR_OF_HIDDEN_NEURONS.add(POSSIBLE_HIDDEN_NEURONS[CURRENT_NR_OF_HIDDEN_NEURONS_INDEX]);
    // NUMBER OF COLUMNS IN THE HIDDEN LAYER GRID
    HIDDEN_COLUMNS [i] = int(sqrt(NR_OF_HIDDEN_NEURONS.get(i)));
    // SIZE OF THE HIDDEN LAYER (IN PIXELS)
    HIDDEN_SIZES   [i] = int(HIDDEN_COLUMNS[i]*HIDDEN_DOT_SIZE+(HIDDEN_COLUMNS[i]-1)*4);
  }

  // CALCULATE POSITIONS FOR DISPLAY
  input_offset_x  = (width-INPUT_SIZE)/2;
  input_offset_y  = height-INPUT_SIZE-20;

  output_offset_x = (width-OUTPUT_WIDTH)/2;
  output_offset_y = 42;

  // CONNECTIONS
  inputCenterTop     = new PVector(input_offset_x+(INPUT_SIZE>>1), input_offset_y);
  outputCenterBottom = new PVector(output_offset_x+(OUTPUT_WIDTH>>1), output_offset_y);
  outputLeftBottom   = new PVector(output_offset_x, output_offset_y);

  int hSpace = 0;
  int vSpace = 0;

  // HIDDEN LAYERS
  if (NR_OF_HIDDEN_LAYERS > 0)
  {
    hSpace = (width - (NR_OF_HIDDEN_LAYERS * HIDDEN_SIZES[0])) / (NR_OF_HIDDEN_LAYERS + 1);
    vSpace = (height - OUTPUT_HEIGHT - HIDDEN_SIZES[0] - INPUT_SIZE - 20) / 2;
    for (int i=0; i<NR_OF_HIDDEN_LAYERS; i++)
    { 
      hidden_offset_x[i]    = i * HIDDEN_SIZES[0] + (i + 1) * hSpace;
      hidden_offset_y[i]    = vSpace + OUTPUT_HEIGHT;
      hiddenCenterTop[i]    = new PVector(hidden_offset_x[i] + (HIDDEN_SIZES[0]>>1), hidden_offset_y[i]);      
      hiddenCenterBottom[i] = new PVector(hidden_offset_x[i] + (HIDDEN_SIZES[0]>>1), hidden_offset_y[i] + HIDDEN_SIZES[0]);
      hiddenCenterLeft[i]   = new PVector(hidden_offset_x[i], hidden_offset_y[i] + (HIDDEN_SIZES[0]>>1));
      hiddenCenterRight[i]  = new PVector(hidden_offset_x[i] + HIDDEN_SIZES[0], hidden_offset_y[i] + (HIDDEN_SIZES[0]>>1));
    }
  }

  // LOAD SIGMOID
  sigmoid = new Sigmoid();   

  // CREATE THE NETWORK
  neuralnet = new Network(NR_OF_INPUT_NEURONS, NR_OF_OUTPUT_NEURONS);  

  nrOfTrainedImages  = 0;
  nrOfTestedImages   = 0;
  nrOfMatchedImages  = 0;
  fSuccessPercentage = 0.0;
  fLastAccuracy      = 0.0;

  // FORCE THE DISPLAY TO REDRAW
  testingNextImage = true;
} // initSettings()